/* Optimizes "zxc -s" output by removing redundant zero loads,
   changing them to register loads when needed, and removing
   double ex de,hl pairs.  Not much of an optimization, but
   better than nothing.  To be used when the Hi-Tech C optimizer
   runs out of memory. */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#define NUMREGS ('l'-'a'+1)
int iszero[NUMREGS];

struct
{
    char name[66];
    int used_ix;
    int used_iy;
} functions[1000];

int numfunctions=0;

char *getline(char *buf)
{
    char op[80],arg1[80];
    int i;
    if(NULL==gets(buf)) return NULL;
    for(i=0;buf[i];i++)
        if(buf[i]=='\t') buf[i]=' ';
    if(*buf==';') return buf;
    if(!strncmp(buf,"push",4)) return buf;
    if(!strcmp(buf,"or a")) return buf;
    if(!strncmp(buf,"ldi",3) || !strncmp(buf,"cpi",3)
     || !strncmp(buf,"outi",4) || !strncmp(buf,"ini",3)
     || !strncmp(buf,"ldd",3) || !strncmp(buf,"cpd",3)
     || !strncmp(buf,"outd",4) || !strncmp(buf,"ind",3))
    {
        iszero['h'-'a']=iszero['l'-'a']=iszero['b'-'a']=iszero['c'-'a']
         = iszero['d'-'a']=iszero['e'-'a']=0;
        return buf;
    }
    if(!strncmp(buf,"ex",2))
    {
        if(!strcmp(buf,"ex de,hl"))
        {
            int sw;
            sw=iszero['h'-'a'];
            iszero['h'-'a']=iszero['d'-'a'];
            iszero['d'-'a']=sw;
            sw=iszero['l'-'a'];
            iszero['l'-'a']=iszero['e'-'a'];
            iszero['e'-'a']=sw;
        }
        else /* exx */
          for(i=0;i<NUMREGS;i++) iszero[i]=0;
        return buf;
    }
    if(!strncmp(buf,"djnz",4))
    {
        iszero['b'-'a']=0;
        return buf;
    }

    if(NULL!=strchr(buf,':') || !strncmp("call",buf,4)
     || !strncmp("jr",buf,2) || !strncmp("jp",buf,2)
     || !strncmp("ret",buf,3))
    {
        for(i=0;i<NUMREGS;i++) iszero[i]=0;
        return buf;
    }
    if(!strncmp("ld ",buf,3) && !strcmp(buf+4,",0")) return buf;
    if(!strncmp("ld ",buf,3) && isalpha(buf[3]) && isalpha(buf[4])
      && buf[5]==',' && buf[3]!='i')
    {
        /* ld mn,... */
        int i;
        unsigned value;
/*        fprintf(stderr,"%s\n",buf); */
        for(i=6;buf[i];i++)
          if(!isdigit(buf[i]))
             goto NotDoubleLoadNumber;
        sscanf(buf+6,"%u",&value);
/*        fprintf(stderr,"[%u]\n",value); */
        if((value&0xFF) == 0)
            iszero[buf[4]-'a']=1;
        else iszero[buf[4]-'a']=0;
        if((value&0xFF00) == 0)
            iszero[buf[3]-'a']=1;
        else iszero[buf[3]-'a']=0;
/*
        for(i=0;i<NUMREGS;i++) if(iszero[i]) fputc(i+'a',stderr);
        fputc('\n',stderr);
*/
        return buf;
    }
NotDoubleLoadNumber:
    *arg1=0;
    *op=0;
    sscanf(buf,"%s %s",op,arg1);
    if(isalpha(*arg1))
    {
        iszero[*arg1-'a']=0;
        if(isalpha(*arg1))
         iszero[arg1[1]-'a']=0;
    }
    return buf;
}

int main()
{
    char buf[1024];
    char buf2[1024];
    char buf3[1024];
    int i;
    int curfunction;
    int used_ix,used_iy;
    int text=0;
    while(NULL!=getline(buf))
    {
        if(!strncmp(buf,"psect ",6))
        {
            if(strncmp(buf+6,"text",4)) text=1;
              else text=0;
            break;
        }
        if(!text) break;
        if(*buf=='_' && NULL!=strchr(buf,':'))
        {
            if(numfunctions)
            {
                functions[numfunctions-1].used_ix=used_ix;
                functions[numfunctions-1].used_iy=used_iy;
            }
            *(strchr(buf,':'))=0;
            strcpy(functions[numfunctions++].name,buf);
            used_ix=used_iy=0;
            break;
        }
        if(!strncmp(buf,"call ",5)) break;
        /* these should really tokenize: there may be variables
           with "ix" and "iy" in them */
        if(!used_ix && NULL!=strstr(buf,"ix")) used_ix=1;
        if(!used_iy && NULL!=strstr(buf,"iy")) used_iy=1;
    }
    rewind(stdin);
    while(NULL!=getline(buf))
    {
        if(!strncmp(buf,"ld ",3) && !strcmp(buf+4,",0"))
        {
            if(iszero[buf[3]-'a'])
            {
                printf(";;*miniopt* removed %s\n",buf);
                continue;
            }
            for(i=0;i<NUMREGS;i++)
                if(iszero[i])
                {
                    printf("ld %c,%c ;;*miniopt* was ld %c,0\n",buf[3],i+'a',buf[3]);
                    break;
                }
            if(i==NUMREGS)
            {
                if(buf[3]=='a') puts("xor a ;;*miniopt* was ld a,0");
                 else puts(buf);
            }
            iszero[buf[3]-'a']=1;
        }
        else
        if(!strcmp(buf,"ex de,hl"))
        {
            if(NULL==getline(buf2))
            {
                puts(buf);
                return;
            }
            if(strcmp(buf2,"ex de,hl"))
            {
                puts(buf);
                puts(buf2);
            }
            else
                puts(";;*miniopt* removed double ex de,hl");

        }
        else puts(buf);
    }
return 0;
}


