#include "zxcc.h"

#define BDOS_DEF
#include "zxbdos.h"
#include "zxcbdos.h"
#include "zxdbdos.h"

#ifdef __MSDOS__
#include <conio.h>
#endif

#define BCD(x) (((x % 10)+16*(x/10)) & 0xFF)

/* Convert time_t to CP/M day count/hours/minutes */
dword cpmtime(time_t t)
{
        long d  = (t / 86400) - 2921;  /* CP/M day 0 is unix day 2921 */
        long h  = (t % 86400) / 3600;  /* Hour, 0-23 */
        long m  = (t % 3600)  / 60;    /* Minute, 0-59 */

        return (d | (BCD(h) << 16) | (BCD(m) << 24));
}


byte get_time(cpm_word b)
{
	time_t t;

	time(&t);
	wr32(b, cpmtime(t));
	
	return (BCD(t % 60));
}


/* Functions to access 24-bit & 32-bit words in memory. These are always
  little-endian. */

void wr24(word addr, dword v)
{
	RAM[addr    ] =  v        & 0xFF;
	RAM[addr + 1] = (v >> 8)  & 0xFF;
	RAM[addr + 2] = (v >> 16) & 0xFF;
}

void wr32(word addr, dword v)
{
        RAM[addr    ] =  v        & 0xFF;
        RAM[addr + 1] = (v >> 8)  & 0xFF;
        RAM[addr + 2] = (v >> 16) & 0xFF;
	RAM[addr + 3] = (v >> 24) & 0xFF;
}

dword rd24(word addr)
{
	register dword rv = RAM[addr + 2];
	
	rv = (rv << 8) | RAM[addr + 1];
	rv = (rv << 8) | RAM[addr];
	return rv;
}


dword rd32(word addr)
{
        register dword rv = RAM[addr + 3];

	rv = (rv << 8) | RAM[addr + 2];
        rv = (rv << 8) | RAM[addr + 1];
        rv = (rv << 8) | RAM[addr];
        return rv;
}

#define peekw(addr) ( (((word)(RAM[addr + 1])) << 8) | RAM[addr])


/* Get / set the program return code. We store this in 'C' form: 0 for
   success, 1-255 for failure. Translate to/from the CP/M form of:

   0x0000-0xFEFF for success
   0xFF00-0xFFFE for failure 

   We also store the actual value so it can be returned

  */

word cpm_errcde(word DE)	
{
	static word real_err = 0;

	if (DE == 0xFFFF) return real_err;
	real_err = DE;

	if      (DE == 0xFF00) cpm_error = 1;
	else if (DE  > 0xFF00) cpm_error = (DE & 0xFF);	
	else		       cpm_error = 0;
	return 0;
}


#ifdef USE_CPMGSX
gsx_byte gsxrd(gsx_word addr)
{
	return RdZ80(addr);
}

void gsxwr(gsx_word addr, gsx_byte value)
{
	WrZ80(addr, value);
}

#endif


void cpmbdos(Z80 *R)
{
	byte *pde  = &RAM[R->DE.W];
	byte *pdma = &RAM[cpm_dma];
	word temp;
	int retv;

/*	Msg("BDOS with C=%02x DE=%04x\n", R->BC.B.l, R->DE.W);  */

	switch(R->BC.B.l)
	{
		case 0: 
		R->PC.W = 0;
		break;

		case 1:		/* Get a character */
#ifdef USE_CPMIO
		retv = cpm_bdos_1();
#else
#ifdef __MSDOS__
		retv = getch(); putch(retv);
#else
		retv = getchar();
#endif
#endif
		if (retv < 0) R->PC.W = 0;
		R->HL.W = retv;
		break;

		case 2:		/* Print a character */
#ifdef USE_CPMIO
		if (cpm_bdos_2(R->DE.B.l)) R->PC.W = 0;	
#else
#ifdef __MSDOS__
		putch(R->DE.B.l);
#else
		putchar(R->DE.B.l);
#endif
#endif
		break;

		case 3:		/* No auxin */
		R->HL.W = 0x1A;
		break;

		case 4:		/* No auxout */
		break;

		case 5:		/* No printer */
		break;

		case 6:		/* Direct console I/O */
		retv = cpm_bdos_6(R->DE.B.l);
		if (retv < 0) R->PC.W = 0;
		R->HL.W = retv;
		break;

		case 7:		/* No auxist */
		case 8:		/* No auxost */
		break;

		case 9:		/* Print a $-terminated string */
#ifdef USE_CPMIO
		if (cpm_bdos_9((char *)pde)) R->PC.W = 0;
#else
		while (RAM[R->DE.W] != '$') 
#ifdef __MSDOS__
			putch(RAM[R->DE.W++]);
#else
			putchar(RAM[R->DE.W++]);	
#endif
#endif
		break;	

		case 0x0A:
		bdos_rdline(R->DE.W, &(R->PC.W));
		break;

		case 0x0B:	/* Console status */
		R->HL.W = 0;	/* No keys pressed */
		break;

		case 0x0C:	/* Get CP/M version */

/* For GENCOM's benefit, claim to be v3.1 */
		
		R->HL.W = 0x31;	/* CP/M v3.1 */
	   /*	R->HL.W = 0x22;	 * CP/M v2.2 */
		break;

		case 0x0D:	/* Re-log discs */
		fcb_reset(); 	
		break;	
	
		case 0x0E:	/* Set default drive */
		R->HL.W = fcb_drive(R->DE.B.l);
		break;

		case 0x0F:	/* Open using FCB */
		R->HL.W = x_fcb_open(pde, pdma);
		break;

		case 0x10:	/* Close using FCB */
		R->HL.W = fcb_close(pde);
		break;

		case 0x11:	/* Find first */
		R->HL.W = fcb_find1(pde, pdma);
		break;

		case 0x12:
		R->HL.W = fcb_find2(pde, pdma);
		break;

		case 0x13:	/* Delete using FCB */
		R->HL.W = fcb_unlink(pde, pdma);
		break;

                case 0x14:      /* Sequential read using FCB */
                R->HL.W = fcb_read(pde, pdma);
                break;

                case 0x15:      /* Sequential write using FCB */
                R->HL.W = fcb_write(pde, pdma);
                break;

		case 0x16:	/* Create using FCB */
		R->HL.W = fcb_creat(pde, pdma);
		break;

		case 0x17:	/* Rename using FCB */
		R->HL.W = fcb_rename(pde, pdma);
		break;

		case 0x18:	/* Get login vector */
		R->HL.W = fcb_logvec();
		break;

		case 0x19:	/* Get default drive */
		R->HL.W = cpm_drive;
		break;

		case 0x1A:	/* Set DMA */
		cpm_dma = R->DE.W;
		break;

		case 0x1B:	/* Get alloc vector */
		fcb_getalv(RAM + 0xFF80, 0x40);
		R->HL.W = 0xFF80;
		break;

		case 0x1C:	/* Make disc R/O */
		R->HL.W = fcb_rodisk();	
		break;

		case 0x1D:	/* Get R/O vector */
		R->HL.W = fcb_rovec();
		break;

		case 0x1E:	/* Set attributes */
		R->HL.W = fcb_chmod(pde, pdma);
		break;

		case 0x1F:	/* Get DPB */
		fcb_getdpb(RAM + 0xFFC0);
		R->HL.W = 0xFFC0;

		case 0x20:	/* Get/set uid */
		R->HL.W = fcb_user(R->DE.B.l);
		break;

                case 0x21:      /* Read a record */
                R->HL.W = fcb_randrd(pde, pdma);
                break;

                case 0x22:      /* Write a record */
                R->HL.W = fcb_randwr(pde, pdma);
                break;

		case 0x23:	/* Get file size */
		R->HL.W = x_fcb_stat(pde);
		break;

		case 0x24:	/* Get file pointer */
		R->HL.W = fcb_tell(pde);
		break;

		case 0x25:
		R->HL.W = fcb_resro(R->DE.W);
		break;

		/* MP/M drive access functions, not implemented */

		case 0x28:	/* Write with 0 fill */
		R->HL.W = fcb_randwz(pde, pdma);
		break;

		/* MP/M record locking functions, not implemented */

		case 0x2C:	/* Set no. of records to read/write */
		R->HL.W = fcb_multirec(R->DE.B.l);
		break;

		case 0x2D:	/* Set error mode */
		err_mode = R->DE.B.l;
		break;

		case 0x2E:
		R->HL.W = fcb_dfree(R->DE.B.l, pdma);

		/* 0x2F: Chain */

		case 0x30:
		R->HL.W = fcb_sync(R->DE.B.l);
		break;	

		case 0x31:
		if (pde[1] == 0xFE) 
		{
			RAM[0xFE9C + *pde] = pde[2];
			RAM[0xFE9D + *pde] = pde[3];
		}
		else if (RAM[R->HL.W + 1] == 0xFF)
		{
			RAM[0xFE9C + *pde] = pde[2];
		}
		else 
		{
			R->HL.B.l = RAM[0xFE9C + *pde];
			R->HL.B.h = RAM[0xFE9D + *pde];
		}
		break;

		case 0x32:
		temp = R->IX.W;		
		R->IX.B.l = 3 * (pde[0] + 1);
		R->AF.B.h = pde[1];
		R->BC.B.l = pde[2];
		R->BC.B.h = pde[3];
		R->DE.B.l = pde[4];
		R->DE.B.h = pde[5];
		R->HL.B.l = pde[6];
		R->HL.B.h = pde[7];
		cpmbios(R);	
		R->IX.W  = temp;
		break;

		case 0x3C:	/* Communicate with RSX */
		R->HL.W = 0;
		break;

		case 0x62:	/* Purge */
		R->HL.W = fcb_purge();
		break;

		case 0x63:	/* Truncate file */
		R->HL.W = fcb_trunc(pde, pdma);
		break;

		case 0x64:	/* Set label */
		R->HL.W = fcb_setlbl(pde, pdma);
		break;

		case 0x65:	/* Get label byte */
		R->HL.W = fcb_getlbl(R->DE.B.l);
		break;

                case 0x66:      /* Get file date */
                R->HL.W = fcb_date(pde);
                break;

		case 0x67:	/* Set password */
		R->HL.W = fcb_setpwd(pde, pdma);
		break;

		case 0x68:	/* Set time of day */
		/* Not advisable to let an emulator play with the clock */
		break;

		case 0x69:	/* Get time of day */
		R->HL.W = get_time(R->DE.W);
		break;	

		case 0x6A:	/* Set default password */
		R->HL.W = fcb_defpwd(pde);
		break;

		case 0x6B:	/* Get serial number */
		memcpy(pde, SERIAL, 6);
		break;

		case 0x6C:	/* 0.03 set error code */
		R->HL.W = cpm_errcde(R->DE.W);
		break;

#ifdef USE_CPMIO
		case 0x6D:	/* Set/get console mode */
		R->HL.W = cpm_bdos_109(R->DE.W);
		break;

		case 0x6E:	/* Set/get string delimiter */
		R->HL.W = cpm_bdos_110(R->DE.B.l);
		break;

		case 0x6F:	/* Send fixed length string to screen */
		if (cpm_bdos_111((char *)RAM + peekw(R->DE.W), 
                                 peekw(R->DE.W + 2)))
			R->PC.W = 0;
		break;

		case 0x70:	/* Send fixed length string to printer */
		break;

		/* 0x71: Strange PCP/M function */
#else
		case 0x6D:	/* Set/get console mode */
		R->HL.W = 0;
		break;

#endif

#ifdef USE_CPMGSX
		case 0x73:	/* GSX */
		R->HL.W = gsx80(gsxrd, gsxwr, R->DE.W);	
		break;	
#endif

		case 0x74:	/* Set date stamp */
		R->HL.W = fcb_sdate(pde, pdma);
		break;

		case 0x98:	/* Parse filename */
		R->HL.W = fcb_parse((char *)RAM + peekw(R->DE.W),
                                    (byte *)RAM + peekw(R->DE.W + 2));
		break;

		default:
#ifdef USE_CPMIO
		cpm_scr_unit();
#endif
#ifdef USE_CPMGSX
	        gsx_deinit();
#endif

		fprintf(stderr,"%s: Unsupported BDOS call %d\n", progname, R->BC.B.l);	
		dump_regs(stderr,R);
		exit(1);
		break;	
	}
	
	R->AF.B.h = R->HL.B.l;
	R->BC.B.h = R->HL.B.h;
}



void cpmbios(Z80 *R)
{
	int func = (R->IX.B.l / 3) - 1;

	switch(func)	/* BIOS function */
	{
		case 1:
		CPURunning = 0;	/* Program termination */
		break;

		case 2:		/* CONST */
#ifdef USE_CPMIO
		R->AF.B.h = cpm_const();
#else 
		R->AF.B.h = cpm_bdos_6(0xFE);
#endif
		break;

		case 3: 	/* CONIN */
#ifdef USE_CPMIO
		R->AF.B.h = cpm_conin();
#else 
		R->AF.B.h = cpm_bdos_6(0xFD);
#endif
		break;

		case 4:		/* CONOUT */
#ifdef USE_CPMIO
		cpm_conout(R->BC.B.l);
#else 
		cpm_bdos_6(R->BC.B.l);
#endif
		break;

		case 20:	/* DEVTBL */
		R->HL.W = 0xFFFF;
		break;

		case 22:	/* DRVTBL */
		R->HL.W = 0xFFFF;
		break;

		case 26:	/* TIME */
		RAM[0xFEF8] = get_time(0xFEF4);
		break;	

		case 30:	/* USERF!!! */
                CPURunning = 0; /* Program termination */
#ifdef USE_CPMIO
		cpm_bdos_110('$');
		cpm_bdos_9("This program has attempted to call USERF, "
                           "which is not implemented\r\n$");
#else
		printf("This program has attempted to call USERF, which "
                       "is not implemented.\n");
#endif
		break;

		default:
#ifdef USE_CPMIO
                cpm_scr_unit();
#endif
#ifdef USE_CPMGSX
                gsx_deinit();
#endif

                fprintf(stderr,"%s: Unsupported BIOS call %d\n", progname, func);
                dump_regs(stderr,R);
                exit(1);
	}
}
